"""implement api to access google sheet
"""
import logging
from typing import Optional, List
import re

from googleapiclient.discovery import Resource

from pysuite.utilities import retry_on_out_of_quota, MAX_RETRY_ATTRIBUTE, SLEEP_ATTRIBUTE

VALID_DIMENSION = {"COLUMNS", "ROWS"}


class Sheets:
    """provide api to operate google spreadsheet. An authenticated google api client is needed.

    :param service: an authorized Google Spreadsheet service client.
    :param max_retry: max number of retry on quota exceeded error. if 0 or less, no retry will be attempted.
    :param sleep: base number of seconds between retries. the sleep time is exponentially increased after each retry.
    """

    def __init__(self, service: Resource, max_retry: int=0, sleep: int=5):
        self._service = service.spreadsheets()
        setattr(self, MAX_RETRY_ATTRIBUTE, max_retry)
        setattr(self, SLEEP_ATTRIBUTE, sleep)

    @retry_on_out_of_quota()
    def download(self, id: str, sheet_range: str, dimension: str= "ROWS", fill_row: bool=False) -> list:
        """download target sheet range by specified dimension. All entries will be considered as strings.

        :param id: id of the target spreadsheet.
        :param sheet_range: range in the target spreadsheet. for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :param dimension: "ROW" or "COLUMNS". If "ROWS", each entry in the output list would be one row in the
          spreadsheet. If "COLUMNS", each entry in the output list would be one column in the spreadsheet.
        :param fill_row: Whether force to return rows with desired number of columns. Google Sheet API ignores trailing
          empty cells by default. By setting this to True, empty strings will be filled in those ignored cells. This
          parameter only works when dimension is "ROWS".
        :return: content of target sheet range in a list of lists.
        """
        if dimension not in VALID_DIMENSION:
            raise ValueError(f"{dimension} is not a valid dimension. expecting {VALID_DIMENSION}.")

        result = self._service.values().get(spreadsheetId=id,
                                            range=sheet_range,
                                            majorDimension=dimension).execute()
        values = result.get('values', [])

        if fill_row and dimension == "ROWS":
            col_counts = get_col_counts_from_range(sheet_range)
            self._fill_rows(values, col_counts)

        return values

    @retry_on_out_of_quota()
    def upload(self, values: list, id: str, range: str) -> None:
        """Upload a list of lists to target sheet range. All entries in the provided list must be serializable.

        :param values: a list of lists of objects that can be converted to str.
        :param id: id of the target spreadsheet
        :param range: range in the target spreadsheet. for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        self.clear(id=id, sheet_range=range)
        body = {"values": values}
        logging.info(f"Updating sheet '{id}' range '{range}'")
        request = self._service.values().update(spreadsheetId=id,
                                                range=range,
                                                valueInputOption="RAW",
                                                body=body)
        result = request.execute()
        msg = f"{result.get('updatedRange')} has been updated ({result.get('updatedRows')} rows " \
              f"and {result.get('updatedColumns')} columns)"
        logging.info(msg)

    @retry_on_out_of_quota()
    def clear(self, id: str, sheet_range: str):
        """remove content in the target sheet range.

        :param id: id of the target spreadsheet
        :param sheet_range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        self._service.values().clear(spreadsheetId=id, range=sheet_range, body={}).execute()

    def read_sheet(self, id: str, sheet_range: str, header=True, dtypes: Optional[dict]=None,
                   columns: Optional[list]=None, fill_row: bool=True):
        """download the target sheet range into a pandas dataframe. this method will fail if pandas cannot be imported.

        :param id: id of the target spreadsheet
        :param sheet_range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :param header: whether first row is used as column names in the output dataframe.
        :param dtypes: a mapping from column name to the type. if not None, type conversions will be applied to columns
          requested in the dictionary.
        :param columns: a list of column names. If not None and `header` is False, this will be used as columns of the
          output dataframe
        :param fill_row: Whether attempt to fill the trailing empty cell with empty strings. This prevents errors when
          the trailing cells in some rows are empty in the sheet. When header is True, this will attempt to fill the
          missing header with _col{i}, where i is the index of the column (starting from 1).
        :return: a pandas dataframe containing target spreadsheet values.
        """
        try:
            import pandas as pd
        except ModuleNotFoundError as e:
            logging.critical("read_sheet() requires pandas.")
            raise e

        if dtypes is not None and not isinstance(dtypes, dict):
            raise TypeError(f"dtypes must be dictionary. got {type(dtypes)}")

        values = self.download(id=id, sheet_range=sheet_range, fill_row=fill_row)
        if values == []:
            return pd.DataFrame()

        if header:
            columns = values.pop(0)
            if fill_row:
                for i in range(len(columns)):
                    if columns[i] == "":
                        columns[i] = f"_col{i+1}"

        df = pd.DataFrame(values, columns=columns)

        if dtypes is not None:
            for col, type in dtypes.items():
                df[col] = df[col].astype(type)

        return df

    def to_sheet(self, df, id: str, sheet_range: str):
        """Upload pandas dataframe to target sheet range. The number of columns must fit the range. More columns or
        fewer columns will both raise exception. The data in the provided dataframe must be serializable.

        :param df: pandas dataframe to be uploaded
        :type df: pandas.DataFrame
        :param id: id of the target spreadsheet
        :param sheet_range: range in the target spreadsheet.  for example, 'sheet!A1:D'. this means selecting from tab "sheet"
          and download column A to D and rows from 1 to the last row with non-empty values.
        :return: None
        """
        values = df.fillna('').values.tolist()
        values.insert(0, list(df.columns))  # insert column names to first row.
        self.upload(values, id=id, range=sheet_range)

    @retry_on_out_of_quota()
    def create_spreadsheet(self, name: str) -> str:
        """create a spreadsheet with requested name.

        :param name: name of the created sheet.
        :return: id of the spreadsheet
        """
        file_metadata = {
            "properties": {"title": name}
        }
        response = self._service.create(body=file_metadata, fields="spreadsheetId").execute()
        return response.get("spreadsheetId")

    @retry_on_out_of_quota()
    def batch_update(self, id: str, body: dict):
        """low level api used to submit a json body to make changes to the specified spreadsheet.

        :param id: id of the target spreadsheet
        :param body: request json
        :return: response from batch upate
        """
        response = self._service.batchUpdate(spreadsheetId=id, body=body).execute()
        return response

    def create_sheet(self, id: str, title: str):
        """create a new sheet with given name in the specified spreadsheet.

        :param id: id of the spreadsheet
        :param title: title of the new sheet
        :return: a dictionary containing information about created sheet, such as sheet id, title, index.
        """
        request = {"requests": [{"addSheet": {"properties": {"title": title}}}]}
        response = self.batch_update(id=id, body=request)
        info = response.get("replies")[0]["addSheet"]["properties"]
        return info

    def delete_sheet(self, id: str, sheet_id: int):
        """delete the specified sheet in the target spreadsheet. You can find sheet_id from URL when you select the
        sheet in the spreadsheet after "gid="

        :param id: id of spreadsheet.
        :param sheet_id: id of sheet
        :return: None
        """
        request = {"requests": [{"deleteSheet": {"sheetId": sheet_id}}]}
        self.batch_update(id=id, body=request)

    def rename_sheet(self, id: str, sheet_id: int, title: str):
        """rename a sheet in target spreadsheet to the new title.

        :param id: id of the target spreadsheet
        :param sheet_id: id of the sheet
        :param title: new title of the sheet
        :return: None
        """
        request = {"requests": [
            {"updateSheetProperties": {"properties": {"sheetId": sheet_id, "title": title}, "fields": "title"}}
        ]
        }
        self.batch_update(id=id, body=request)

    def _fill_rows(self, rows: List[list], col_counts: int):
        for row in rows:
            if len(row) < col_counts:
                row.extend(['']*(col_counts - len(row)))


def get_column_number(col: str) -> int:
    """Convert spreadsheet column numbers to integer.

    :example:

    >>> get_column_number('A')  # 1
    >>> get_column_number('AA') # 27
    >>> get_column_number('ZY') # 701

    :param col: upper case spreadsheet column
    :return: index of the column starting from 1.
    """
    result = 0
    l = len(col)
    for i in range(l):
        result += (ord(col[l-i-1]) - 64) * (26 ** i)

    return result


def get_col_counts_from_range(sheet_range: str) -> int:
    """Calculate the number of columns in the given range.

    :example:

    >>> get_col_counts_from_range("test!A1:A")  # 1
    >>> get_col_counts_from_range("test!A1:D")  # 4
    >>> get_col_counts_from_range("test!AA2:AZ")  # 26

    :param sheet_range: a string representation of sheet range. For example, "test_sheet!A1:D"
    :return: the number of columns contained in the range.
    """
    columns = sheet_range.upper().split("!")[1]  # get the columns in letter, such as "A1:D"
    from_column, to_column = columns.split(":")
    pattern = re.compile("^([A-Z]+)", re.IGNORECASE)
    from_column = pattern.search(from_column).group(0)
    to_column = pattern.search(to_column).group(0)
    counts = get_column_number(to_column) - get_column_number(from_column) + 1
    return counts



"""classes used to authenticate credentials and create service for Google Suite Apps
"""
from typing import Union, Optional
from pathlib import Path, PosixPath
import json
import logging

from googleapiclient.discovery import build
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from google.cloud import vision as gv
from google.cloud import storage
from google.auth._default import load_credentials_from_file


SCOPES = {
    "drive": "https://www.googleapis.com/auth/drive",
    "sheets": "https://www.googleapis.com/auth/spreadsheets",
    "gmail": "https://www.googleapis.com/auth/gmail.compose",
    "vision": "https://www.googleapis.com/auth/cloud-vision",
    "storage": None,
}

CLOUD_SERVICES = {"vision", "storage"}

DEFAULT_VERSIONS = {
    "drive": "v3",
    "sheets": "v4",
    "gmail": "v1",
    "vision": "v1",
    "storage": None,
}


class Authentication:
    """read from credential file and token file and authenticate with Google service for requested services. if token
    file does not exists, confirmation is needed from browser prompt and the token file will be created. You can pass
    a list of services or one service.
    """
    def __init__(self, credential: Union[PosixPath, str], services: Union[list, str],
                 token: Optional[Union[PosixPath, str]] = None):
        self._token_path = Path(token) if token is not None else None  # can be None if requesting google cloud service
        self._credential_path = Path(credential)
        self._services = self._get_services(services)
        self._scopes = self._get_scopes()
        self._credential = self.load_credential()
        self.refresh()

    def load_credential(self) -> Credentials:
        """load credential json file needed to authenticate Google Suite Apps. If token file does not exists,
        confirmation is needed from browser prompt and the token file will be created.

        :param credential: path to the credential json file.
        :return: a Credential object
        """
        if not self.is_google_cloud:
            if self._token_path is None:
                raise ValueError(f"token is required for {self._services}.")

            if not Path(self._token_path).exists():
                return self._load_credential_from_file(self._credential_path)  # pragma: no cover

            with open(self._token_path, 'r') as f:
                token_json = json.load(f)

            with open(self._credential_path, 'r') as f:
                try:
                    cred_json = json.load(f)["installed"]
                except KeyError:
                    raise KeyError("'installed' does not exist in credential file. please check the format")

            try:
                credential = Credentials(token=token_json["token"],
                                         refresh_token=token_json["refresh_token"],
                                         token_uri=cred_json["token_uri"],
                                         client_id=cred_json["client_id"],
                                         client_secret=cred_json["client_secret"],
                                         scopes=self._scopes,
                                         )
            except KeyError as e:
                logging.critical("missing key value in credential or token file")
                raise e
        else:
            credential, _ = load_credentials_from_file(str(self._credential_path))

        return credential

    def _load_credential_from_file(self, file_path: PosixPath) -> Credentials:
        """load credential json file and open web browser for confirmation.

        :param file_path: path to the credential json file.
        :return: a Credential object
        """
        if self._services is None:
            raise ValueError("service must not be None when token file does not exists")

        flow = InstalledAppFlow.from_client_secrets_file(file_path, self._scopes)
        credential = flow.run_local_server(port=9999)
        return credential

    def refresh(self):
        """refresh token if not valid or has expired. In addition token file is overwritten.

        :return: None
        """
        if not self.is_google_cloud:
            if not self._credential.valid:
                if self._credential.expired and self._credential.refresh_token:  # pragma: no cover
                    self._credential.refresh(Request())

            self.write_token()
        else:
            logging.warning("Google cloud service do not require refresh of token.")

    def write_token(self):
        token_json = {
            "token": self._credential.token,
            "refresh_token": self._credential.refresh_token
        }
        with open(self._token_path, 'w') as token:
            json.dump(token_json, token)

    def get_service_client(self, service: Optional[str]=None, version: Optional[str]=None):
        """get a service object for requested service. This service must be within authorized scope set up at
        initiation stage.

        :param service: type of service, "drive" or "sheets". If None and self._services has more than 1 items, an
          exception will be raised.
        :param version: version of target service. if None, default version will be used. it varies with service.
        :return: a service object used to access API for that service.
        """
        if service is None:
            if len(self._services) > 1:
                raise ValueError(f"service cannot be inferred. the authorized services are {self._services}")

            service = self._services[0]
        elif service not in self._services:
            raise ValueError(f"service {service} is not among authorized services: {self._services}")

        if version is None:
            version = DEFAULT_VERSIONS[service]

        if service not in CLOUD_SERVICES:
            return build(service, version, credentials=self._credential, cache_discovery=True)
        elif service == "vision":
            return gv.ImageAnnotatorClient(credentials=self._credential)
        elif service == "storage":
            return storage.Client(credentials=self._credential)
        else:
            # Won't reach here
            raise ValueError(f"Invalid service: {service}. This is an implementation error.")  # pragma: no cover

    def _get_scopes(self) -> list:
        try:
            scopes = [SCOPES[service] for service in self._services]
            return scopes
        except KeyError as e:
            logging.critical(f"{self._services} is not a valid service. expecting {SCOPES.keys()}")
            raise e

    def _get_services(self, services: Union[list, str]) -> list:
        if isinstance(services, str):
            services = [services]
        if not set(services).issubset(SCOPES.keys()):
            raise ValueError(f"invalid services. got {services}, expecting {SCOPES.keys()}")

        if set(services).intersection(CLOUD_SERVICES):
            diff = set(services).difference(CLOUD_SERVICES)
            if diff:
                raise ValueError(f"Google cloud services {CLOUD_SERVICES} cannot be mixed with non cloud services. "
                                 f"Found {diff}")

        return services

    @property
    def is_google_cloud(self):
        return set(self._services).issubset(CLOUD_SERVICES)



"""Implement api to access google storage API
"""
from pathlib import PosixPath, Path
from typing import Union

from google.cloud.storage.client import Client, Bucket

GS_HEADER = "gs://"


class Storage:
    """Class to interact with Google Storage API.

    :param service: an authorized Google Storage service client.
    """

    def __init__(self, service: Client):
        self._service = service

    def upload(self, from_object: Union[str, PosixPath], to_object: str):
        """Upload a file or a folder to google storage. If `from_object` is a folder, this method will
        upload it recursively.

        :param from_object: Path to the local file or folder to be uploaded.
        :param to_object: Target Google storage object location. If `from_object` is a file, this will be a file. If
          `from_object` is a folder, this will be a folder. This is a string that looks like "gs://xxxxx"
        :return: None
        """
        from_object: PosixPath = Path(from_object).resolve()
        if not from_object.exists():
            raise IOError(f"{from_object} does not exist.")

        _bucket, _gs_object = self._split_gs_object(to_object)
        bucket = self.get_bucket(bucket_name=_bucket)
        if from_object.is_file():
            blob = bucket.blob(_gs_object)
            blob.upload_from_filename(str(from_object))
        else:
            for _from, _to in _add_folder_tree_to_new_base_dir(from_object, _gs_object):
                if _from.is_file():
                    blob = bucket.blob(_to)
                    blob.upload_from_filename(str(_from))

    def download(self, from_object: str, to_object: Union[str, PosixPath]):
        """Download target Google storage file or folder to local. If `from_object` is a folder, this method will
        download it recursively.

        :param from_object: Target Google storage path to be downloaded. This is a string that looks like "gs://xxxx"
        :param to_object: Path to the local file or folder. If `from_object` is a file, this will be a file. If
          `from_object` is a folder, this will be a folder.
        :return: None
        """
        to_object: PosixPath = Path(to_object)
        blobs = list(self.list(target_object=from_object))
        if len(blobs) == 1:
            # No way we can tell if it's a folder or file, always consider it as file
            blobs[0].download_to_filename(str(to_object))
        else:
            for blob in blobs:
                _to_file = to_object / blob.name
                _to_file.parent.mkdir(parents=True, exist_ok=True)
                blob.download_to_filename(str(_to_file))

    def remove(self, target_object: str):
        """Remove target Google storage file or folder. If `target_object` is a folder, this will remove it recursively.

        :param target_object: Target Google storage file or folder. This is a string that looks like "gs://xxxx"
        :return: None
        """
        _bucket, _ = self._split_gs_object(target_object)
        bucket = self.get_bucket(bucket_name=_bucket)
        bucket.delete_blobs(blobs=list(self.list(target_object=target_object)))

    def copy(self, from_object: str, to_object: str):
        """Copy Google storage file or folder from one location to another. If `from_object` is a folder, this will
        copy it recursively.

        :param from_object: Source Google storage file or folder. This is a string that looks like "gs://xxxx"
        :param to_object: Destination Google storage file or folder. This is a string that looks like "gs://xxxx"
        :return: None
        """
        _src_bucket, _src_gs_object = self._split_gs_object(from_object)
        _dest_bucket, _dest_prefix = self._split_gs_object(to_object)
        src_bucket = self.get_bucket(_src_bucket)
        dest_bucket = self.get_bucket(_dest_bucket)
        blobs = list(src_bucket.list_blobs(prefix=_src_gs_object))
        if len(blobs) == 1:
            src_bucket.copy_blob(blobs[0], dest_bucket, _dest_prefix)
        else:
            _src_prefix_len = len(_src_gs_object)
            for blob in blobs:
                name = blob.name
                _dest_gs_object = _dest_prefix + name[_src_prefix_len:]
                src_bucket.copy_blob(blob, dest_bucket, _dest_gs_object)

    def list(self, target_object: str):
        """Search Google storage target location and return an iterator. This iterator generates all files under the
        target location. If the target is a single file, the iterator only one object.

        :param target_object: Target Google storage location. This could be a file or a folder. This is a string that
          looks like "gs://xxxxx"
        :return: An iterator that iterates over the target location. Each item is a Blob object.
        """
        _bucket, _gs_object = self._split_gs_object(target_object=target_object)
        bucket = self.get_bucket(bucket_name=_bucket)
        blob_iterator = bucket.list_blobs(prefix=_gs_object)
        return blob_iterator

    def create_bucket(self, bucket_name: str) -> Bucket:
        """Create a bucket in Google Storage.

        :param bucket_name: The name of the Google storage bucket.
        :return: None
        """
        return self._service.create_bucket(bucket_name)

    def get_bucket(self, bucket_name: str) -> Bucket:
        """Get a Bucket object for the target Google storage bucket.

        :param bucket_name: The name of the target bucket.
        :return: A Bucket object for the target bucket.
        """
        return self._service.get_bucket(bucket_name)

    def remove_bucket(self, bucket_name: str, force: bool = False):
        """Remove the target bucket.

        :param bucket_name: Target bucket name.
        :param force: Whether force remove the target bucket. If True, even if the bucket is not empty, it will be
          removed. Default is False.
        :return:
        """
        bucket = self._service.get_bucket(bucket_name)
        bucket.delete(force=force)

    def _split_gs_object(self, target_object: str) -> (str, str):
        """Split a string that looks like "gs://bucket_name/object/path" into bucket name and object path. If it is not
        a valid gs path, an ValueError will be raised.

        :param target_object: Target google storage path.
        :return: A tuple of string. (bucket name, object path)
        """
        if not is_gcs_uri(target_uri=target_object):
            raise ValueError(f"{target_object} is not a valid gs object.")

        bucket, object_path = target_object[len(GS_HEADER):].split("/", 1)
        return bucket, object_path


def is_gcs_uri(target_uri: str):
    return isinstance(target_uri, str) and target_uri.startswith(GS_HEADER)


def _add_folder_tree_to_new_base_dir(from_path: PosixPath, to_path: str) -> (PosixPath, str):
    """Construct Google storage folder tree based on local folder tree so that the hierarchy is maintained.

    :param from_path: Path to a local folder.
    :param to_path: Path to the target Google storage folder.
    :return: Iterate and yield tuples of (local file, corresponding Google storage file Path)
    """
    folder_tree = from_path.rglob("*")
    for f in folder_tree:
        relative_path = f.relative_to(from_path)
        yield f, to_path + "/" + str(relative_path)



"""Implement api to access google vision API
"""
import logging
import json
from pathlib import PosixPath
from typing import Union, Optional, List
import warnings

from google.cloud import vision as gv
from google.cloud.vision_v1.types import AnnotateImageResponse
from google.cloud.vision_v1.types.image_annotator import BatchAnnotateImagesResponse
from google.cloud.vision_v1 import types, ImageAnnotatorClient
from google.api_core.operation import Operation

from pysuite.storage import is_gcs_uri


class Vision:
    """Class to interact with Google Vision API.

    :param service: an authorized Google Vision service client.
    """

    def __init__(self, service: ImageAnnotatorClient):
        self._service = service
        self._requests = []

    @staticmethod
    def load_image(image_path: Union[str, PosixPath]) -> gv.Image:
        """Load a local image as Image class that can be used to submit image annotation requests.

        :param image_path: Path to the image file.
        :return: Loaded Image object from the target file
        """
        with open(image_path, "rb") as f:
            image = gv.Image(content=f.read())
            return image

    def add_request(self, image_path: Union[str, PosixPath], methods: Union[List[str], str]):
        """Add a request to annotate a local or GCS image. Multiple annotation methods can be added at the same time.
        The request will not be immediately submitted. This method is useful for `batch_annotate_image` and
        `async_annotate_image`.

        :example:

        >>> vision.add_request("/my/image.png", methods=["LABEL_DETECTION", "test_detection"])

        :example:

        >>> vision.add_request("gs://my_gcs_bucket/image/path.jpg", methods=["LABEL_DETECTION", "test_detection"])

        :param image_path: Local or GCS path to the image file.
        :param methods: A list of strings representing supported annotation methods. Please view
          google.cloud.vision_v1.types.Feature.Type for all supported methods. They are case-insensitive.
        :return: None
        """
        request = self._create_request(image_path, methods)
        self._requests.append(request)

    def annotate_image(
            self, image_path: Union[str, PosixPath], methods: Union[List[str], str]
    ) -> AnnotateImageResponse:
        """Submit a request to annotate a local image using specified methods.

        :param image_path: Path to the image file.
        :param methods: A list of strings representing supported annotation methods. Please view
          google.cloud.vision_v1.types.Feature.Type for all supported methods. They are case-insensitive.
        :return: An AnnotateImageResponse object with annotated content.
        """
        request = self._create_request(image_path, methods)

        response = self._service.annotate_image(request=request)
        return response

    def batch_annotate_image(self) -> Optional[BatchAnnotateImagesResponse]:
        """Submit the prepared requests to annotate images and return a response with annotated content. You must first
        call `add_request` to prepare the configurations. If no configurations were prepared, this method will return
        None.

        :return: An BatchAnnotateImagesResponse object with annotated content. Or None if no requests were prepared.
        """
        if not self._requests:
            warnings.warn("No requests was prepared", UserWarning)
            return

        response = self._service.batch_annotate_images(requests=self._requests)
        return response

    def async_annotate_image(self, output_gcs_uri: str, batch_size: int = 0) -> Optional[Operation]:
        """Annotate images asynchronously and place output in Google Cloud Storage in batches.

        :param output_gcs_uri: Target output location on Google Cloud Storage.
        :param batch_size: Maximum number request processed in each batch. If there are 10 requests submitted together,
          and `batch_size` is 2, there will be 5 output files (batches) created. Default is 0. Meaning only create one
          batch.
        :return: An Operation object. You can use it to wait until the process is completed.
        """
        if not self._requests:
            warnings.warn("No requests was prepared", UserWarning)
            return None

        output_config = {
            "gcs_destination": {"uri": output_gcs_uri},
            "batch_size": batch_size
        }

        response = self._service.async_batch_annotate_images(requests=self._requests, output_config=output_config)
        return response

    @staticmethod
    def _translate_method(method: str):
        """Translate string method to corresponding feature type in google.cloud.vision_v1.types.Feature.Type. This is
        case insensitive. If no such type is implemented, a NotImplementedError will be raised.

        :param method: A string representation of feature type implemented in google vision.
        :return: The corresponding attributes of feature type.
        """
        try:
            return getattr(types.Feature.Type, method.upper())
        except AttributeError as e:
            raise NotImplementedError(f"Cannot find requested method {method}.") from e

    @staticmethod
    def _create_request(image_path: Union[str, PosixPath], methods: Union[List[str], str]) -> dict:
        if isinstance(methods, str):
            methods = [methods]
        features = []
        for method in methods:
            features.append({"type_": Vision._translate_method(method)})

        if is_gcs_uri(image_path):
            # Google storage image uri is provided
            image = {"source": {"image_uri": image_path}}
        else:
            # Local image path is provided
            image = Vision.load_image(image_path)

        request = {
            "image": image,
            "features": features
        }
        return request

    @staticmethod
    def to_json(response: Union[AnnotateImageResponse, BatchAnnotateImagesResponse]) -> dict:
        """Convert possible image responses to dictionary. If it's not a supported response, a type error will be
        raised.

        :param response: A response returned from `annotate_image` or `batch_annotate_image`.
        :return: A dictionary containing annotated contents.
        """
        if isinstance(response, AnnotateImageResponse):
            annotated = json.loads(types.image_annotator.AnnotateImageResponse.to_json(response))
        elif isinstance(response, BatchAnnotateImagesResponse):
            annotated = json.loads(types.image_annotator.BatchAnnotateImagesResponse.to_json(response))
        else:
            raise TypeError(f"Invalid type of response. Expecting AnnotateImageResponse or BatchAnnotateImagesResponse."
                            f"Got {type(response)}")

        return annotated



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path
from setuptools import setup, find_packages

here = Path(__file__).resolve().parent

def get_version():
    init_file = here / "pysuite" / "__init__.py"
    with open(init_file, 'r') as f:
        for l in f.readlines():
            if l.startswith("__version__"):
                version = eval(l.strip().split(" ")[-1])
                return version

    raise RuntimeError("version not found")

NAME = 'pysuite'
DESCRIPTION = "A data scientist's toolbox for Google Suite Apps"
URL = 'https://staftermath.github.io/pysuite/index.html'
EMAIL = 'gwengww@gmail.com'
AUTHOR = 'Weiwen Gu'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = get_version()

REQUIRED = [
    "google-api-python-client>=1.7.8",
    "google-api-core>=1.31.1",
    "google-auth>=1.20.1",
    "pyparsing>=2.4.7, <3.0.0",
    "google-auth-httplib2>=0.0.4",
    "google-auth-oauthlib>=0.4.1",
    "googleapis-common-protos>=1.6",
    "google-cloud-vision>=2.4.2",
    "google-cloud-storage>=1.42.3"
]

try:
    with open(here / 'README.md', encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {'__version__': VERSION}

setup(
    name=NAME,
    version=about['__version__'],
    license='bsd-3-clause',
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    download_url = f'https://github.com/staftermath/pysuite/archive/{VERSION}.tar.gz',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    install_requires=REQUIRED,
    include_package_data=True,
)


import functools
import re
import warnings
import time
from typing import Optional

from googleapiclient.errors import HttpError

MAX_RETRY_ATTRIBUTE = "max_retry"
SLEEP_ATTRIBUTE = "sleep"


def retry_on_type_and_msg(exception: Exception, msg_pattern: Optional[str] = None):
    def wrapper(method):
        @functools.wraps(wrapped=method)
        def wrapped_function(self, *args, **kwargs):
            max_retry = getattr(self, MAX_RETRY_ATTRIBUTE, 0)
            max_retry = max(max_retry, 0)
            sleep = getattr(self, SLEEP_ATTRIBUTE, 5)
            if sleep < 0:
                raise AttributeError(f"{SLEEP_ATTRIBUTE} must be positive. Got {sleep}")

            pattern = None
            if msg_pattern is not None:
                pattern = re.compile(msg_pattern)
            while True:
                max_retry -= 1
                try:
                    result = method(self, *args, **kwargs)
                    return result
                except exception as e:
                    if max_retry < 0 or pattern is not None and not pattern.match(str(e)):
                        # max retry reached or pattern does not match.
                        raise e

                    warnings.warn(f"handled exception {e}. remaining retry: {max_retry}", UserWarning)
                    time.sleep(sleep)
                    sleep = sleep*2
                    continue

        return wrapped_function

    return wrapper


def retry_on_out_of_quota():
    """A decorator to give wrapped function ability to retry on quota exceeded related HttpError raise by Google API.
    It only works on class method and requires "max_retry" and "sleep" attribute in the class. If `max_retry` is
    non-positive, no retry will be attempt. `sleep` is the base number of seconds between consecutive retries. The number
    of wait seconds will double after each sleep.

    :return:
    """
    return retry_on_type_and_msg(HttpError, msg_pattern=".*(User Rate Limit Exceeded|Quota exceeded)+.*")



"""implement api to access gmail
"""
import logging
from pathlib import PosixPath
from base64 import urlsafe_b64encode
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from typing import Union, Optional, List

from googleapiclient.discovery import Resource


class GMail:
    """A class containing methods to interact with Gmail APIs such as sending emails

    :param service: an authorized GMail service client.
    """

    def __init__(self, service: Resource):
        self._service = service.users().messages()

    def compose(self, sender: str, to: Union[str, list], cc: Optional[Union[str, list]]=None,
                bcc: Optional[Union[str, list]]=None, body: Optional[str]=None, subject: Optional[str]=None,
                local_files: Optional[Union[str, PosixPath]]=None,
                gdrive_ids: Optional[List[str]]=None,
                user_id: Optional[str]=None,
                is_html: bool=True) -> dict:
        """Compose an email and send through gmail api.

        :param sender: email of the sender
        :param to: a string of one email or list of emails of the targets
        :param cc: a string of one email or list of emails of cc'ed targets
        :param bcc: a string of one email or list of emails of bcc'ed targets
        :param body: main content in email body.
        :param subject: subject line
        :param local_files: list of files to be attached in the email
        :param gdrive_ids: list of gdrive ids attached in the email
        :param user_id: identification of user. If None, it sender email will be used.
        :param is_html: whether body is send in html format. If False, plain text format will be used.
        :return: a dictionary of response. containing 'id' (id of the email). 'threadId' (id of the email thread) and
          'labelIds' (list of label ids)
        """
        msg = self._create_message_skeleton(subject=subject,
                                            sender=sender,
                                            to=to,
                                            cc=cc,
                                            bcc=bcc)
        msg = self._attach_body(msg, body, ids=gdrive_ids, is_html=is_html)
        msg = self._attach_local_files(msg=msg, files=local_files)
        if user_id is None:
            user_id = sender
        response = self._send(user_id=user_id, msg=msg)
        return response

    def _attach_local_files(self, msg: MIMEBase, files: Optional[List[Union[str, PosixPath]]]) -> MIMEBase:
        """Attach a list of local files to msg. If files is None, no changes will be made.

        :param msg: a MIMEBase object to attach files with.
        :param files: list of local files to be attached.
        :return: a MIMEBase object with files attached.
        """
        if files is None:
            return msg

        for file in files:
            file = PosixPath(file)
            with open(file, 'rb') as fp:
                attachment = MIMEBase("application", "octet-stream")
                attachment.set_payload(fp.read())
                attachment.add_header('Content-Disposition', 'attachment', filename=file.name)
                msg.attach(attachment)
        return msg

    def _attach_body(self, msg: MIMEBase, body: Optional[str], ids: Optional[List[str]], is_html: bool) -> MIMEBase:
        """Attach email body to the msg. If gdrive ids are provided, attach gdrive file hyperlinks in the body. The
        format of the email can be specified to 'html' or 'plain'.

        :param msg: a MIMEBase object.
        :param body: body of the email.
        :param ids: a list of gdrive file ids.
        :param is_html: whether email should be send in html format or plain text format
        :return: a msg with body attached.
        """
        if body is not None:
            body_mime = MIMEText(body, 'html' if is_html else 'plain')
            msg.attach(body_mime)

        msg = self._attach_gdrive_files(msg, ids)
        return msg

    def _attach_gdrive_files(self, msg: MIMEBase, ids: Optional[List[str]]) -> MIMEBase:
        """Attach a list of gdrive files to the msg body in forms of hyper links. If ids is None, no attachment will be
        added.

        :param msg: a MIMEBase object
        :param ids: list of gdrive ids.
        :return: a MIMEBase object with gdrive link attached
        """
        if ids is None:
            return msg

        gdrive_section = """<p>GDrive Attachment</p>\n<ul>\n\t{attached}\n</ul>"""
        link_template = "<li>https://drive.google.com/file/d/{id}</li>"
        attached = []
        for id in ids:
            attached.append(link_template.format(id=id))
        gdrive_section = gdrive_section.format(attached="\n\t".join(attached))
        gdrive_html = MIMEText(gdrive_section, "html")
        msg.attach(gdrive_html)
        return msg

    def _create_message_skeleton(self, sender: str, to: Union[str, list], cc: Optional[Union[str, list]]=None,
                                 bcc: Optional[Union[str, list]]=None, subject: Optional[str]=None) -> MIMEMultipart:
        message = MIMEMultipart()
        message['from'] = sender
        message['to'] = self._format_recipients(to)
        if subject is not None:
            message['subject'] = subject
        if cc is not None:
            message['cc'] = self._format_recipients(cc)
        if bcc is not None:
            message['bcc'] = self._format_recipients(cc)
        return message

    def _send(self, user_id: str, msg: MIMEBase) -> dict:
        """Send composed email through API.

        :param user_id: displayed user id.
        :param msg: A composed MIMEBase object.
        :return: dictionary of response
        """
        body = {'raw': urlsafe_b64encode(msg.as_bytes()).decode(),
                'payload': {'mimeType': 'text/html'}}
        response = self._service.send(userId=user_id, body=body).execute()
        logging.debug(response)
        return response

    def _format_recipients(self, recipients: Union[str, list]) -> str:
        """Convert a list of emails to a string accepted by gmail API.

        :param recipients: list of emails.
        :return: a string representing all recipients.
        """
        if not isinstance(recipients, list) and not isinstance(recipients, str):
            raise TypeError(f"recipients must be str or list type. Got {type(recipients)}")

        if isinstance(recipients, str):
            return recipients

        return ", ".join(recipients)
